// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file MeshPartitioner.h */

#ifndef AMDIS_MESH_PARTITIONER_H
#define AMDIS_MESH_PARTITIONER_H

#include <map>
#include <set>
#include <mpi.h>

#include "AMDiS_fwd.h"
#include "Mesh.h"
#include "parallel/MpiHelper.h"
#include "parallel/ParallelDofMapping.h"


namespace AMDiS {

  using namespace std;

  enum PartitionMode {
    INITIAL = 0,          // initial partitioning of a unpartitioned mesh 
    ADAPTIVE_REPART = 1,  // repartitioning of a adaptively refined mesh
    REFINE_PART = 2       // quality improvement of the current partitioning
  };


  /**
   * Abstract class for mesh partitioning. This class provides only a function
   * for a random initial partitioning. A concrete partition must override the 
   * functions \ref MeshPartitioner::partition and 
   * \ref MeshPartitioner::createPartitionMap.
   */
  class MeshPartitioner
  {
  public:
    MeshPartitioner(MPI::Intracomm *comm)
      : mpiComm(comm),
	mesh(NULL),
	boxPartitioning(false),
	mapLocalGlobal(NULL)
    {}

    virtual ~MeshPartitioner() {}

    /// Creates an initial paritioning of the AMDiS mesh. This partitioning
    /// can be arbitrary, the only requirement is that each macro element
    /// must be uniquely assign to a rank.
    virtual void createInitialPartitioning();

    /** \brief
     * Function the takes a weighted set of macro elements and returns a 
     * macro mesh partitioning. This function is virtual and must be implemented
     * for a specific algorithm or an external partitioning library.
     *
     * \param[in]  elemWeights   Maps to each macro element in rank's subdomain
     *                           a weight, which is usually the number of leaf
     *                           elements in this macro element.
     * \param[in]  mode          Most external partitioning libraries can make
     *                           a difference whether we want to create a
     *                           first partitioning or we alread have created
     *                           one using this library but due to some mesh
     *                           adaptivity we want to repartition the mesh. In
     *                           the later case, the libraries also consider the
     *                           time for redistribution of the new partitioning.
     * \return     Returns a boolean value if the partitioning algorithm created
     *             a correct partitioning. If it is so, the partitioning is 
     *             stored in \ref elementInRank and \ref partitionMap.
     */
    virtual bool partition(map<int, double> &elemWeights,
			   PartitionMode mode = INITIAL) = 0;

    virtual void createPartitionMap(map<int, int>& partitionMap) = 0;

    /// Write partitioner state to disk.
    void serialize(ostream &out);

    /// Read partitioner state from disk.
    void deserialize(istream &in);

    void setMesh(Mesh *m)
    {
      mesh = m;
    }

    void setBoxPartitioning(bool b)
    {
      boxPartitioning = b;
    }

#if 0
    void setLocalGlobalDofMap(map<DegreeOfFreedom, DegreeOfFreedom> *m)
#else
      void setLocalGlobalDofMap(map<DegreeOfFreedom, MultiIndex> *m)
#endif
    {
      mapLocalGlobal = m;
    }

    Mesh *getMesh()
    {
      return mesh;
    }

    map<int, bool>& getElementInRank()
    {
      return elementInRank;
    }

    map<int, vector<int> >& getRecvElements()
    {
      return recvElements;
    }

    map<int, vector<int> >& getSendElements()
    {
      return sendElements;
    }
    
    /// After mesh repartition this function returns true if the mesh must be 
    /// redistributed on at least one rank.
    bool meshChanged()
    {
      int nChanges = recvElements.size() + sendElements.size();
      mpi::globalAdd(nChanges);

      return static_cast<bool>(nChanges);
    }

  protected:
    /// Pointer to the MPI communicator the mesh partitioner should make use of.
    MPI::Intracomm *mpiComm;

    /// Pointer to the AMDiS mesh.
    Mesh *mesh;

    /// The mesh partitioner can be used in to different modes, the standard
    /// mode and the so called "box partitioning". The standard mode assigns
    /// macro elements to ranks. If box partitioning is enabled, which makes
    /// only sence if the macro mesh results from meshconv's "lego mesher",
    /// then in 2D boxed (2 macro elements) and in 3D cubes (6 macro 
    /// elements) are assigned as a uniion to ranks. 
    bool boxPartitioning;

    /// In box partitioning mode this map stores for each box number the set
    /// of macro element indices the box consists of.
    map<int, std::set<int> > boxSplitting;
    
    /// In box partitioning mode this map stores to each box number the set
    /// of neighbouring boxes.
    map<int, std::set<int> > boxNeighbours;

    /// Is the reverse of the map \ref boxSplitting. Thus, it stores for each
    /// macro element index the box number it belongs to.
    map<int, int> elInBox;

    map<DegreeOfFreedom, MultiIndex> *mapLocalGlobal;

    map<int, vector<int> > elNeighbours;

    /// Maps to each macro element index (or box index in box 
    /// partitioning mode) if it is in rank's partition or not.
    map<int, bool> elementInRank;

    /// Maps to each macro element index (or box index in box
    /// partitiong mode) the rank number the element belongs to.
    map<int, int> partitionMap;

    /// After mesh repartitioning these maps stores which elements are communicated
    /// from this rank to other ranks.
    map<int, vector<int> > recvElements, sendElements;
  };
}

#endif
